class EpubError(Exception):
    """
    Base exception class for all exceptions raised by the openepub library.
    """

    pass


class InvalidFileError(EpubError):
    """
    Indicates the file is invalid and does not conform to the EPUB specification.
    """

    pass


class MalformedPackage(EpubError):
    pass


class DRMProtected(EpubError):
    """
    Indicates the EPUB content is protected by Digital Rights Management.
    """

    pass



from unittest import TestCase

import bs4

from openepub import Epub
from openepub.common import collect_text


class CollectTextTestCase(TestCase):
    def test_multiple_spans_in_p(self):
        """
        Test to make sure that if there are span elements within block elements
        their spaces aren't trimmed from the ends.

        Targets a fix for github issue #2.
        """

        source = (
            """<p><span>Odpovědi na tyto a mnohé další otázky hledá v novém"""
            """ senzačním životopise nazvaném </span><span>Život a lži Albu"""
            """se Brumbála</span><span> Rita Holoubková. Exkluzivní rozhovo"""
            """r s ní přináší Betty Braithwaiteová na straně 13.</span></p>"""
        )

        expected = (
            """Odpovědi na tyto a mnohé další otázky hledá v novém senzační"""
            """m životopise nazvaném Život a lži Albuse Brumbála Rita Holou"""
            """bková. Exkluzivní rozhovor s ní přináší Betty Braithwaiteová"""
            """ na straně 13.\n"""
        )

        soup = bs4.BeautifulSoup(source.encode("utf-8"), "html.parser")
        collected = collect_text(soup)

        self.assertEqual(collected, expected)

    def test_romeo_and_juliet(self):
        """
        Test to make sure we are reading Romeo & Juliet correctly.
        """

        epub = Epub("test/mock/prj_gb_romeo_and_juliet.epub")
        with open("test/mock/prj_gb_romeo_and_juliet.txt", "r") as f:
            expected = f.read()
        collected = epub.get_text()

        self.assertEqual(collected, expected)

    def test_right_to_left(self):
        """
        Test to make sure a right-to-left language (Farsi) book is being read
        correctly.
        """

        epub = Epub("test/mock/prj_gb_five_selected_short_stories_farsi.epub")
        with open("test/mock/prj_gb_five_selected_short_stories_farsi.txt", "r") as f:
            expected = f.read()
        collected = epub.get_text()

        self.assertEqual(collected, expected)

    def test_multi_encoding_match(self):
        """
        Test to make sure that in the presence of the chardet library that Unicode
        texts are not incorrectly decoded as ISO-8859-1.

        Targets a fix for github issue #3.
        """
        try:
            import chardet
        except ImportError:
            self.skipTest("Requires the chardet library to test correctness")

        source = "<p>Trompie die hoërskoolseun</p>".encode("utf-8")
        utf_expected = "Trompie die hoërskoolseun\n"
        iso_expected = "Trompie die hoÃ«rskoolseun\n"

        soup = bs4.BeautifulSoup(source, "html.parser")
        collected = collect_text(soup)
        self.assertEqual(
            collected,
            iso_expected,
            "Expect BeautifulSoup to decode incorrectly if left with defaults",
        )

        soup = bs4.BeautifulSoup(source, "html.parser", from_encoding="utf-8")
        collected = collect_text(soup)
        self.assertEqual(
            collected,
            utf_expected,
            "Expect BeautifulSoup to decode correctly if told to use Unicode",
        )

    def test_iso_encoding(self):
        """
        Verify that ISO encoded text is picked up correctly even with a unicode
        prompt to Beautiful Soup
        """

        source = b"<p>\xc6sir</p>"
        expected = "Æsir\n"

        soup = bs4.BeautifulSoup(source, "html.parser", from_encoding="utf-8")
        collected = collect_text(soup)
        self.assertEqual(collected, expected)

    def test_multi_encoding_epub(self):
        """
        Test to make sure that in the presence of the chardet library that Unicode
        texts are not incorrectly decoded as ISO-8859-1.

        Targets a fix for github issue #3.
        """
        try:
            import chardet
        except ImportError:
            self.skipTest("Requires the chardet library to test correctness")

        epub = Epub("test/mock/hola_multiencoding_match.epub")
        expected = "Trompie die hoërskoolseun"
        collected = epub.get_text()

        self.assertEqual(collected, expected)

    def test_multi_encoding_epub_without_chardet(self):
        """
        Test to make sure that in the absence of the chardet library that Unicode
        texts are not incorrectly decoded as ISO-8859-1.

        Targets a fix for github issue #3.
        """
        try:
            import chardet

            self.skipTest("Shouldn't have the chardet library for this test")
        except ImportError:
            pass

        epub = Epub("test/mock/hola_multiencoding_match.epub")
        expected = "Trompie die hoërskoolseun"
        collected = epub.get_text()

        self.assertEqual(collected, expected)



from bs4 import NavigableString
import re


def aslist(arg):
    if isinstance(arg, list):
        return arg
    if isinstance(arg, dict):
        return [arg]
    return list(arg)


def is_block_tag(tag_name):
    return tag_name in [
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "p",
        "li",
        "tr",
        "br",
        "div",
    ]


def collect_text(soup, within_block=False):
    text = ""
    if isinstance(soup, NavigableString):
        text = soup.text
        text = re.sub(r"\s+", r" ", text)
    elif hasattr(soup, "children"):
        for child in soup.children:
            if is_block_tag(child.name):
                text += collect_text(child, within_block=True) + "\n\n"
            else:
                text += collect_text(child, within_block)
        if not within_block:
            text = "\n".join(line.strip() for line in text.splitlines())
    return text



import os
import typing
import xml.parsers.expat as xml_expat
import xmltodict
import zipfile

import openepub as oe
from openepub.common import aslist


class Epub:
    def __init__(self, path=None, stream=None):
        if path is None and stream is None or path is not None and stream is not None:
            raise ValueError("One of path or stream must be specified.")

        if path is None:
            pass
        elif isinstance(path, (str, bytes, os.PathLike)):
            if not os.path.exists(path):
                raise oe.InvalidFileError("Specified path does not exist.")
            path_or_stream = path
        else:
            raise TypeError("path argument must be path-like.")

        if stream is None:
            pass
        elif hasattr(stream, "read"):
            path_or_stream = stream
        else:
            raise TypeError("stream argument must be file-like.")

        if not zipfile.is_zipfile(path_or_stream):
            raise oe.InvalidFileError("File specified is not an EPUB.")

        self.zip = zipfile.ZipFile(path_or_stream, mode="r")
        self.root = zipfile.Path(self.zip)

        self.package_paths = self._read_package_paths()
        if not self.package_paths:
            raise oe.InvalidFileError("File specified is not an EPUB.")

    def _read_package_paths(self) -> list:
        package_paths = []
        try:
            content_bytes = (self.root / "META-INF" / "container.xml").read_bytes()
            container = xmltodict.parse(content_bytes.decode("utf-8", "ignore"))
            for rootfile in aslist(container["container"]["rootfiles"]["rootfile"]):
                path = self.root.joinpath(rootfile["@full-path"])
                if not path.exists():
                    raise ValueError("EPUB rootfile has an invalid path.")
                package_paths.append(path)
        except (
            xml_expat.ExpatError,
            FileNotFoundError,
            KeyError,
            TypeError,
            ValueError,
        ):
            return []
        return package_paths

    def iterpackages(self) -> typing.Iterator["oe.EpubPackage"]:
        for path in self.package_paths:
            yield oe.EpubPackage(epub=self, path=path)

    def get_text(self):
        texts = []
        for package in self.iterpackages():
            for item in package.iterspine():
                texts.append(item.get_text())
        return "\n".join(texts).strip()



import bs4
import xml.parsers.expat as xml_expat
import xmltodict
import zipfile

import openepub as oe
from openepub.common import collect_text


class EpubItem:
    def __init__(self, package: "oe.EpubPackage", path: zipfile.Path, id: str):
        self.package = package
        self.path = path
        self.id = id


class EpubXHTMLItem(EpubItem):
    def __init__(self, package: "oe.EpubPackage", path: zipfile.Path, id: str):
        super().__init__(package, path, id)
        self.content_bytes = self.path.read_bytes()
        try:
            self.content_bytes.decode("utf-8", "strict")
        except UnicodeDecodeError:
            try:
                xmltodict.parse(self.content_bytes.decode("utf-8", "ignore"))
            except xml_expat.ExpatError:
                raise oe.DRMProtected("This content is protected by DRM.") from None
        self.soup = bs4.BeautifulSoup(
            self.content_bytes,
            "html.parser",
            from_encoding="utf-8",
        )

    def get_text(self):
        if self.soup.find("head") and self.soup.find("body"):
            text_soup = self.soup.find("body")
        else:
            text_soup = self.soup
        return collect_text(text_soup)
